#-------------------------------------------------------------------------------
# PACKAGE : Bio::Tools::Blast::Run::Webblast.pm
# PURPOSE : To run a Blast analysis on an NCBI server and save the results locally.
# AUTHOR  : Alex Dong Li (ali@genet.sickkids.on.ca) - original webblast version.
#           Steve A. Chervitz (sac@genome.stanford.edu) - this modularized version.
# CREATED : 4 May 1998
# STATUS  : Alpha
# REVISION: $Id: Webblast.pm,v 1.1.1.1 1998/12/11 15:24:24 birney Exp $
# HISTORY :
#
# MODIFIED:
#  1.0, 5 Jun 1998, sac: (initial release)
#    A decent amount of Alex's original code (version 0.92) remains, but 
#    has been extensively modified. Search for comments marked with "SAC:" 
#    for details of my modifications. Generally, my modifications 
#    were as follows:
#   
#  	 -- moved code from a stand-alone script into a module.
#  	 -- segmented script code into logical procedures.
#        -- automatically configure the path for postclient.pl.
#  	 -- works with Bio::Seq.pm objects instead of files.
#  	 -- added support for Blast1 and the beginnings of support for PSI Blast.
#  	 -- changed print() calls to print STDERR (only if $MONITOR is true).
#  	 -- changed system("rm") calls to unlink().
#  	 -- using 'my' instead of 'local' for subroutine-scope vars.
#  	 -- general cleaning up (got it to run with use strict).
#  	 -- changed `date` calls to $Util->date_format().
#        -- added leading underscores to private methods.
#        -- added advanced option W (word size).
#        -- uses Bio::Tools::Blast::HTML.pm for stripping HTML.
#        -- assorted changes in _removeHTMLtags().
#        -- added miscellaneous comments and POD.
#
# For additional modification notes and the latest version, visit the distribution site:
#    http://bio.perl.org/Projects/Blast
#
#
#   Alex's NOTES FROM webblast 0.92: 
#     changes made to blast files:
#  	 1) delete reference lines and a few junk lines at the beginning of the files
#  	 2) add time/date to search..... ...done line
#  	 3) change the relative URLs of entries to abosulute URLs
#     NEED TO DO:
#     retrieve gif files by getclient--for graphical overview
#     03/05/98: remove .seq and .v2 from output filenames
#     02/10/98: -html option controls more now: -html=ON|off|both
#     02/07/98: add a counter for successful blasting
#     01/24/98: now program can take files in fasta or plain text format
#     01/21/98: add new varible $pathOfPostclient
#  	 so that outside users can easily modify it
#  
# For documentation, run this module through pod2html
# (preferably from Perl v5.004 or better).
#
# Copyright (c) 1998 Alex Dong Li/Steve A. Chervitz. All Rights Reserved.
#           This module is free software; you can redistribute it and/or 
#           modify it under the same terms as Perl itself.
#-------------------------------------------------------------------------------

package Bio::Tools::Blast::Run::Webblast;

use Bio::Root::Global    qw(:devel);
use Bio::Root::Utilities qw(:obj);
use Bio::Tools::Blast::HTML qw(&strip_html);
use Carp;

@ISA        = qw(Exporter);
@EXPORT     = qw();
@EXPORT_OK  = qw(&blast_remote @Blast_dbp_remote @Blast_dbn_remote 
		 @Blast_matrix_remote);
%EXPORT_TAGS = ( std => [qw(&blast_remote @Blast_dbp_remote @Blast_dbn_remote 
			    @Blast_matrix_remote)] );

use strict;
use vars qw( $ID $VERSION $revision);

$ID       = 'Bio::Tools::Blast::Run::Webblast';
$VERSION  = 1.1; 
$revision = '$Id: Webblast.pm,v 1.1.1.1 1998/12/11 15:24:24 birney Exp $';  #'

## POD Documentation:

=head1 NAME

Bio::Tools::Blast::Run::Webblast.pm - Bioperl module for running Blast analyses remotely.

=head1 SYNOPSIS

This module is wrapped by the B<Bio::Tools::Blast.pm> module
and is intended for use via that module. This module can be used
separately if desired:

    use Bio::Tools::Blast::Run::Webblast qw(&blast_remote);
 
    &blast_remote($blast_object, %runParam);

See L<blast_remote>() for a description of available parameters.

This module also exports the following arrays:

 @Blast_dbp_remote     - Available Blast peptide sequence databases
 @Blast_dbn_remote     - Available Blast nucleotide sequence databases
 @Blast_matrix_remote  - Available substitution scoring matrices

=head1 INSTALLATION

This module is included with the central Bioperl distribution:

   http://bio.perl.org/Core/Latest
   ftp://bio.perl.org/pub/DIST

Follow the installation instructions included in the README file.

=head1 DESCRIPTION

Bio::Tools::Blast::Run::Webblast.pm executes a Blast request on an NCBI server 
by mimicking a web (http) client.
This module exports a single function L<blast_remote>(). See the docs for that method
for detailed information. 

Webblast.pm is derived from the stand alone B<webblast>
script by Alex Dong Li:

    http://www.genet.sickkids.on.ca/bioinfo_resources/software.html#webblast

=head1 DEPENDENCIES

B<Bio::Tools::Blast::Run::Webblast.pm> is used by B<Bio::Tools::Blast.pm>
and requires B<postclient.pl> for submitting requests Blast
server. This script is included with the distribution package. 
B<postclient.pl> should be located in the same directory as 
Bio::Tools::Blast::Run::Webblast.pm.

B<Bio::Tools::Blast::Run::Webblast.pm> also uses B<Bio::Seq.pm>
objects. The development of this module is somewhat linked with
Bio::Tools::Blast.pm module and should be updated along with that
module.

=head1 BUGS

=over 0

=item * Support for other Blast servers.

While in theory this module can be used to run a Blast on any remote server, 
it is currently specialized for the NCBI (or NCBI-conforming) servers.
The main source of this dependency is the configuration of options sent 
to the server (see L<_blast>()). There may also be some dependency in the processing
of the results files returned from the server, although I have tried to remove 
these (sac).

=item * PSI-Blast support.

PSI-Blast is not supported. Using a C<-version =E<gt> 'psi'> parameter will 
appear to execute properly but in the resulting file, only the header is obtained. 

=item * Reply by e-mail.

Reply by e-mail is supported but will generate an error since there is no
associated file containing the Blast results.
E-mail reply is probably not appropriate in this module context
since we are typically parsing the results immediately after the run
and we therefore need access to the file. Still, e-mail could be used
if necessary.

=back

=head1 FEEDBACK

=head2 Mailing Lists 

User feedback is an integral part of the evolution of this and other Bioperl modules.
Send your comments and suggestions preferably to one of the Bioperl mailing lists.
Your participation is much appreciated.

    vsns-bcd-perl@lists.uni-bielefeld.de          - General discussion
    vsns-bcd-perl-guts@lists.uni-bielefeld.de     - Technically-oriented discussion
    http://bio.perl.org/MailList.html             - About the mailing lists

=head2 Reporting Bugs

Report bugs to the Bioperl bug tracking system to help us keep track the bugs and 
their resolution. Bug reports can be submitted via email or the web:

    bioperl-bugs@bio.perl.org                   
    http://bio.perl.org/bioperl-bugs/           

=head1 AUTHOR

Steve A. Chervitz, sac@genome.stanford.edu

=head1 VERSION

Bio::Tools::Blast::Run::Webblast.pm, 1.1

=head1 TODO

=over 4

=item * Handle running PSI-Blasts.

=item * Retrieve gif files by getclient (for graphical overview).

=item * Bring up-to-date with latest changes in the forms used by the NCBI servers.

=item * Enable blast_remote() to handle lists of sequence filenames in addition to Bio::Seq.pm objects.

=back


=head1 SEE ALSO

 Bio::Tools::Blast.pm                    - Blast object.
 Bio::Tools::Blast::Run::postclient.pl   - Script for accessing remote server.
 Bio::Tools::Blast::Run::LocalBlast.pm   - Utility module for running Blasts locally.
 Bio::Tools::Blast::HTML.pm              - Blast HTML-formating utility class.
 Bio::Seq.pm                             - Biosequence object  


 http://bio.perl.org/Projects/modules.html  - Online module documentation
 http://bio.perl.org/Projects/Blast/        - Bioperl Blast Project     
 http://bio.perl.org/                       - Bioperl Project Homepage


=head1 COPYRIGHT

Copyright (c) 1998 Alex Dong Li/Steve A. Chervitz. All Rights Reserved.
This module is free software; you can redistribute it and/or 
modify it under the same terms as Perl itself.

=cut


#
##
###
#### END of main POD documentation.
###
##
#'


## SAC: 
## Automatically determine the path for postclient.pl
## (relative to the location of webblast)
my $pathOfPostclient='';
my $path = __FILE__;
if($path =~ /(.+\b)Webblast\.pm/) {
    $pathOfPostclient = $1.'postclient.pl';
} else {
    croak "\nWebblast can't set path for postclient.pl.\n";
}

# SAC: grouped database names.
# May want to put these sorts of things in a
# Blast::Resources.pm type of module.
use vars qw(@Blast_dbp_remote @Blast_dbn_remote @Blast_matrix_remote);

# Peptide datasets
@Blast_dbp_remote = qw(nr month swissprot yeast ecoli kabat pdb alu );

# Nucleotide datasets
@Blast_dbn_remote = qw(nr month swissprot yeast ecoli kabat pdb alu 
		       dbest dbsts htgs vector mito epd gss
		       est_mouse est_human est_others );

# Not necessarily an exhaustive list.
@Blast_matrix_remote  = qw(BLOSUM62 PAM40 PAM120 PAM250 IDENTITY);

# PRIVATE VARIABLES
# -------------------

# SAC: Consolidated the various default values.
# These are the defaults specified by the NCBI server.
my %_default = (
		'expect'                => 10,  
		'expectPSI'             => 0.01,  
		'descriptions'          => 100,
		'alignments'            => 100,
		'queryGeneticCode'      => 1,
		'advancedOptionG'       => 11,  # gap creation
		'advancedOptionE'       => 1,   # gap extension
		'advancedOptionQ'       => -3,  # mismatch penalty,  blastn only
		'advancedOptionR'       => 1,   # match reward,      blastn only
		'advancedOptionW'       => 11,  # word size (11 blastn, 3 others)
		'resultReturnProtocol'  => 'www',   # or 'email'
		'minimumLength'         => 30,
		'minimumLengthProtein'  => 10,
		'maximumLength'         => 100000, 
		'maximumLengthProtein'  => 50000, 
		'blastServerURL1'       => 'www.ncbi.nlm.nih.gov/cgi-bin/BLAST/nph-blast',
		'blastServerURL2'       => 'www.ncbi.nlm.nih.gov/cgi-bin/BLAST/nph-newblast',
		'blastServerURLpsi'     => 'www.ncbi.nlm.nih.gov/cgi-bin/BLAST/nph-psi_blast',
		'entryServerURLroot'    => 'www.ncbi.nlm.nih.gov',
		'imageMapURL'           => 'www.ncbi.nlm.nih.gov/cgi-bin/imagemap/BLAST/',
  # sequenceFormat no longer an issue since we now work with seq objects.
		#'sequenceFormat'        => 'GCG', 
  # notAsk is no longer an option since we are trying to automate this procedure
  # at a higher level.
		#'notAsk'                => 'on',    
		);

my $program                 ='blastp';
my $version                 = 2;
my $gappedAlignmentOption   ='GAPPED_ALIGNMENT=is_set';
my $gappedAlignmentFlag     ='on';
my $database                ='nr';
my $extension               = $program;
my $expect                  = $_default{'expect'};      
my $expectPSI               = $_default{'expectPSI'};      
my $filter                  ='default';             # protein-protein searches only
my $matrix                  ='BLOSUM62';            # scoring matrix 
my $NCBI_giOption           ='';                    # default is off
my $NCBI_giFlag             ='off';                 # default is off
my $descriptions            = $_default{'descriptions'};
my $alignments              = $_default{'alignments'};
my $graphicalOverviewOption =''; # or 'OVERVIEW=is_set';     # not available for email
my $graphicalOverviewFlag   ='off';
my $queryGeneticCode        = $_default{'queryGeneticCode'};

## All advanced options are left undefined unless specified by user.
my $advancedOptionG         = ''; 
my $advancedOptionE         = ''; 
my $advancedOptionQ         = ''; 
my $advancedOptionR         = ''; 
my $advancedOptionW         = ''; 

my $blastServerURL          = '';
my $resultReturnProtocol    = $_default{'resultReturnProtocol'};
my $resultTypeOption        = '';       # null for plain text or html=is_set
my $resultTypeInHTMLflag    ='on';
#$imageMapURL               =  $_default{'imageMapURL'};
#$imageURL=$imageURLdefault ='www.ncbi.nlm.nih.gov/BLAST/';
my $minimumLength           = $_default{'minimumLength'};
my $maximumLength           = $_default{'maximumLength'};
# this is the root of server for retrieving record, might be different from blastServer
my $entryServerURLroot      = $_default{'entryServerURLroot'};
my $emailAddress            = '';
my $currentDate             = '';
my @queue                   = ();  # list of sequence objects

# SAC: new private, class vars
my @_outFiles               = ();
my $_out_dir                = '';    # where to save the Blast output files.
my $_blastObj               = undef;
my $_errmsg                 = '';

# SAC: defunct options and variables:
#my $notAsk             =  $_default{'notAsk'};
#my $sequenceFormat     = $_default{'sequenceFormat'};
#my $queue='';          #A string of tab delimited clone names.
#my $currentTime = '';
#my $username = getpwuid($<);            # get userid
#my @username   = split(//, $username);  # convert the first case into upper
#$username[0] =~ tr/a-z/A-Z/;
#$username    = join("", @username);


######################  BEGIN FUNCTIONS  ########################

=head1 APPENDIX

Methods beginning with a leading underscore are considered private
and are intended for internal use by this module. They are
B<not> considered part of the public interface and are described here
for documentation purposes only.


=head2 blast_remote

 Usage     : @files = blast_remote( $blast_object,  %namedParameters);
           : This method is exported.
 Purpose   : Run a remote Blast analysis on one or more sequences.
 Returns   : Array containing a list of filenames of the Blast reports.
 Argument  : First argument should be a Bio::Tools::Blast.pm object reference.
           : This object is primarily used for error reporting
           : Remaining arguments are named parameters: 
           : (PARAMETER TAGS CAN BE UPPER OR LOWER CASE).
           :
           :   -SEQS       => ref to an array of Bio::Seq.pm objects. 
           :   -PROG       => name of blast program (blastp, blastx, etc.)
           :   -VERSION    => blast version (1 or 2)
           :   -DATABASE   => name of database (see below.)
           :   -EXPECT     => expect value cutoff
           :   -EXPECT_PSI => expect value for inclusion in PSI-BLAST iteration 1 
           :   -FILTER     => sequence complexity filter ('default' or 'none')
           :   -MATRIX     => substitution scoring matrix (blast1 only for NCBI server)
           :   -NCBI_GI    => 'on' or 'off'
           :   -DESCR      => integer, number of on-line descriptions (V, 100)
           :   -ALIGN      => integer, number of alignments (B, 100)
           :   -GRAPH      => 'on' or 'off' (graphical overview not yet supported)
           :   -GAP        => 'on' or 'off'
           :   -GEN_CODE   => integer for special genetic code (see below) blastx only
           :   -HTML       => 'on' or 'off' or 'both'
           :   -SERVER     => blast server to use (default is NCBI Blast2)
           :   -MAX_LEN    => max query sequence length to blast
           :   -MIN_LEN    => min query sequence length to blast
           :   -OUT_DIR    => output directory to store blast result files
           :   -GAP_CREATE => gap creation penalty (G, 5) 
           :   -GAP_EXTEND => gap extension penalty (E, 2)
           :   -MISMATCH   => mismatch penalty (q, -3)  (blastn only)
           :   -MATCH      => match reward (r, 1)       (blastn only)
           :   -WORD       => word size (W, 11 for blastn, 3 for all others)
           :
 Throws    : Exception if:
           :   - Cannot obtain parameters by calling _rearrange() on the
           :     first argument, which should be a Bio::Tools::Blast.pm object ref.
           :   - No sequences are provided.
           :   - Sequence type is incompatible with Blast program type.
           :   - Database name is not one of the valid names.
           :   - Supplied e-mail address looks invalid.
 Comments  :
  -------------------------------------------------------------
  Available programs: blastn, blastx, dbest, blastp, tblastn, tblastx
  Program versions: 1, 2. (Support for PSI-Blasts is incomplete).

  -------------------------------------------------------------
  Available databases are: nr, month, swissprot, dbest, dbsts, 
        est_mouse, est_human, est_others, pdb, vector, kabat,
        mito, alu, epd, yeast, ecoli, gss, htgs.
 
    These are exported by this module in the @Blast_dbp_remote
    and @Blast_dbn_remote arrays.
  -------------------------------------------------------------
  Available Genetic Codes are (blastx only): 
        (1) Standard                    (2) Vertebrate Mitochondrial
        (3) Yeast Mitochondrial         (4) Mold Mitochondrial; ... 
        (5) Invertebrate Mitochondrial  (6) Ciliate Nuclear; ...
        (9) Echinoderm Mitochondrial    (10) Euplotid Nuclear
        (11) Bacterial                  (12) Alternative Yeast Nuclear
        (13) Ascidian Mitochondrial     (14) Flatworm Mitochondrial
        (15) Blepharisma Macronuclear
 
  -------------------------------------------------------------
  Available substitution scoring matrices (NCBI):
      BLOSUM62, PAM40, PAM120, PAM250, IDENTITY.
      Others members of the BLOSUM and PAM family of matrices
      may be available as well.
      These are exported by this module in the @Blast_matrix_remote array.

      Note that certain combinations of matrices and gap creation/extension
      penalties are disallowed (E.g., PAM250 will work with 12/2 but not 11/1).
 --------------------------------------------------------------
   Limited values for gap creation and extension are supported for 
   blastp, blastx, tblastn.  Some supported and suggested values are:

  Creation     Extension

     10             1
     10             2
     11             1
      8             2
      9             2
  -------------------------------------------------------------
  Available sequence complexity filters:
       SEG, SEG+XNU, XNU, dust, none.
 

See Also : _set_options(), _adjust_options(), _validate_options(), _blast(), B<Bio::Tools::Blast.pm>

=cut

#-----------------
sub blast_remote {
#-----------------
    my ($bobj, %param) = @_;
    
    my ($seq_a);
    eval { 
	# _rearrange() is an instance method of Bio::Root::Object.pm and is
	# inherited by the Blast object.
	($seq_a) = $bobj->_rearrange([qw(SEQS)], %param);
    };
    if($@) {
	# Ideally we should use can(), requires Perl 5.004.
	croak "Can't run remote BLAST: failed to obtain parameters using _rearrange().\n".
	    "Make sure that the first argument is a Bio::Tools::Blast.pm object.\n";
    }

    ref($seq_a) eq 'ARRAY' or
	$bobj->throw("Can't get sequences to be Blasted: No sequence data.");

    if( not @queue = @$seq_a) {
	$bobj->throw("Can't Blast: No sequences supplied.");
    }

    @_outFiles = ();

    $_blastObj = $bobj;
    &_set_options(%param);
    &_adjust_options();
    &_validate_options();
    &_blast();

    return @_outFiles;
}

#----------------
sub _set_options {
#----------------
# SAC: Processing function args instead of command-line args
#      Most have the same names as Alex's. Some 
    my (@param) = @_;
    my ($prog, $vers, $db, $exp, $exp_psi, $filt, $mat, $gi, $descr, $aln, 
	$graph, $gap, $gcode, $email, $html, $server, $maxl, $minl, 
	$gap_c, $gap_e, $mmatch, $match, $word, $dir)
	= $_blastObj->_rearrange(
			 [qw(PROG VERSION DATABASE EXPECT EXPECT_PSI FILTER MATRIX 
			  NCBI_GI DESCR ALIGN GRAPH GAP GEN_CODE EMAIL HTML SERVER 
			  MAX_LEN MIN_LEN GAP_CREATE GAP_EXTEND MISMATCH MATCH 
			  WORD OUT_DIR)], @param);
		      
    if($prog) {
	$program = $prog;
	# SAC: commented out advanced option setting. Current strategy
	#      is to set them as needed by the user. Otherwise, let server
	#      pick the defaults for us.
	if ($prog =~ m/^blastx$/i) {
#	    $advancedOptionG = $_default{'advancedOptionG'} = 10;
#	    $advancedOptionE = $_default{'advancedOptionE'} = 1;
	}
	# reset default if submit protein seq.
	if ($prog =~ m/^blastp$/i || $prog=~m/^tblastn$/i) {
#	    $advancedOptionG = $_default{'advancedOptionG'} = 10;
#	    $advancedOptionE = $_default{'advancedOptionE'} = 1;
	    $minimumLength   = $_default{'minimumLengthProtein'};
	    $maximumLength   = $_default{'maximumLengthProtein'};   
	}
    }
    if($vers)    { $version = $vers;  }
    if($gap)     { $gappedAlignmentFlag = $gap;  }
    if($db)      { $database = lc($db);  }
    if($exp)     { $expect = $exp;  }
    if($exp_psi) { $expectPSI = $exp_psi;  }
    if($filt)    { $filter = $filt;   }
    if($mat)     { $matrix = $mat;   }
    if($gi)      { $NCBI_giFlag = $gi;   }
    if($descr)   { $descriptions = $descr;   }
    if($aln)     { $alignments = $aln;   }
    if($graph)   { $graphicalOverviewFlag = $graph;   }
    if($gcode)   { $queryGeneticCode = $gcode; }
    if($gap_c)   { $advancedOptionG = $gap_c; }
    if($gap_e)   { $advancedOptionE = $gap_e; }
    if($mmatch)  { $advancedOptionQ = $mmatch; }
    if($match)   { $advancedOptionR = $match; }
    if($word)    { $advancedOptionW = $word; }
    if($email)   { $resultReturnProtocol="email";
		   $emailAddress = $email; }
    if($server)  { $blastServerURL = $server;}
    if($minl)    { $minimumLength = $minl; } # if match minlength
    if($maxl)    { $maximumLength = $maxl; } # if match maxlength
    if($dir)     { $_out_dir = $dir; }

    if($html) {
	# HTML issue is a bit tricky. Blast.pm requires a non-HTML version
	# for parsing. So if the report is to be parsed, we need to make sure
	# to give Blast.pm the correct version of the results.
	# If HTML is 'on', then only the HTML version is saved and we
	# can't parse. (The Blast object may still try to parse it if its mode is
	# set for parsing, but it will fail). 
	if($html !~ /on|both/i) {
	    $html = 'on';
	}
	$resultTypeInHTMLflag = $html; 
    } else { 
	$resultTypeInHTMLflag = 'off'; 
    }
}


#--------------------
sub _adjust_options {
#--------------------
# reset some varibles depending on the flags
# SAC: fixed bug with gappedAlignmentFlag
#      default is gapped. Need to specify ungapped.
    if ($gappedAlignmentFlag=~m/^on$/i) {
	$gappedAlignmentOption="";
    } else {
	$gappedAlignmentOption="UNGAPPED_ALIGNMENT=is_set";
    }
    
    if ($NCBI_giFlag=~m/^on$/i) {
	$NCBI_giOption="NCBI_GI=is_set";
    } else {
	$NCBI_giOption="";
    }
    
    if ($graphicalOverviewFlag=~m/^on$/i) {
	$graphicalOverviewOption="OVERVIEW=is_set";
    } else {
	$graphicalOverviewOption="";
    }
    
    # off means plain text
    if ($resultTypeInHTMLflag=~m/^off$/i) {
	$resultTypeOption="";         # or html
# with html tags
    } else {
	$resultTypeOption="HTML=is_set";         # or html
    }
    
    
    if ($resultReturnProtocol ne "www") {
	#check if email is in right format here
	if ($emailAddress !~m/\w+@\w+\.\w+/) {
	    $_blastObj->throw("Email address may be wrong. Please check it and try again.");
	}
    } else {
	$currentDate = $Util->date_format('full');
    }
    
    # SAC: select which Blast server to use depending on $version
    # (if a server has not already been selected).
    if(not $blastServerURL) {
	local($_) = $version;
	SWITCH: {
	    /1/i   && do{ $blastServerURL = $_default{'blastServerURL1'}; last SWITCH;};
	    /2/i   && do{ $blastServerURL = $_default{'blastServerURL2'}; last SWITCH;};
	    /psi/i && do{ $blastServerURL = $_default{'blastServerURLpsi'}; last SWITCH;};
	}
    }
}


#----------------------
sub _validate_options {
#----------------------
# some error checks.
# SAC: Throwing exceptions if
#       - can't access postclient.pl
#       - invalid program type
#       - invalid database name

# test to see if the program is executable
    if (!-x "$pathOfPostclient") {
	## SAC: Added extra info to the die call.
	$_blastObj->throw("Can't access postclient.pl: $pathOfPostclient.");
    }
    
# if graphical is on, I need getclient to retrieve the .gif file!
#if ( $graphicalOverviewFlag eq "off" && !-x "/home/ali/script/getclient.pl") {
#    die("\n\nI cannot access getclient.pl, which is needed for graphs. I quit!\a\n\n");
#}

# if program is not of one available program, quit
    # SAC: simplified this conditional
    if (!($program=~m/^t?blast[npx]$/i or $program=~m/^dbest$/i)) {
	$_blastObj->throw("Invalid program name: $program",
			  "It must be one of blastn, blastx, blastp, tblastn, tblastx, or dbest!");
    }
    
# if database is not of one available databases, quit
    # SAC: simplified this conditional and error message
    # SAC: Only checking against nucleotide db's since they are a superset.
    if ( not(grep $database eq $_, @Blast_dbn_remote)) {
	my $valid = join(', ', @Blast_dbn_remote);
	$_blastObj->throw("Invalid Blast database name: $database",
			  "It must be one of $valid");
    }

    # SAC: we are now loading the queue via Blast.pm.

#    if(not @queue) { 
#	 if ($queue ne '') {
#	     $queue=~s/^\t//;
#	     @queue=split(/\t/, $queue);
#	 } else {
#	     # do all files in current dir!
#	     if ( $notAsk eq "off") {
#		 &All_files;
#	     }
#	     $temp=`ls`;
#	     @queue=split(/\n/, $temp);
#	 }
#    }
    
# if $program is not dbest, use program as output file extension
    if ($program !~m/^dbest$/i) {
	$extension  = $program;  # SAC: note that program case is not changed.
# otherwise use blastn as program, dbest as database name and output file extension
    } else {
	$program="blastn";
	$database="dbest";
	$extension="dbest";
    }
}

#------------
sub _blast {
#------------
# work on all sequences one by one
# SAC: Some alterations in the basic plan of this function.
    my ($seq, $typeDo, $advancedOptions);
    my ($id, $outputFileNameTemp, $outputFileName, $baseFileName, $options);
    my $countSuccessfulBlasting = 0;

    my $count = 0;
    foreach $seq (@queue) {

	$count++;
	# Not catching fatal error if the sequence is the wrong type
	# (serious error).
	next unless &_validate_seq($seq);
	
	# add >seq_name to make seq in fasta format
	my $sequenceInFastaFormat= $seq->layout('Fasta');
	$sequenceInFastaFormat =~ s/\n$//;
	
	# concatenate advanced options into a string, depending on differnt blast type
	# options -q and -r must be lower case!
	# SAC: setting options only if they are defined.
	$advancedOptions  = $advancedOptionG ?  "-G $advancedOptionG" : '';
	$advancedOptions .= $advancedOptionE ? " -E $advancedOptionE" : '';
	$advancedOptions .= $advancedOptionW ? " -W $advancedOptionW" : '';
	if ($program=~m/^blastn$/i) {
	    $advancedOptions .= $advancedOptionQ ? " -q $advancedOptionQ" : '';
	    $advancedOptions .= $advancedOptionR ? " -r $advancedOptionR" : '';
	}
	
	# get temp filename and final filename
	$id = $seq->id;
	$id =~ s/\.\w+?$//;  # trim off the extension if id is a filename.
	$baseFileName = $_out_dir. $id . "." . $extension."$version.$database"; # SAC: new var.
	$outputFileNameTemp = $baseFileName. ".temp.html";
	$outputFileName     = $baseFileName. ".html";

	# test to see if I can write temp file. If not, do nothing:
	if (open(OUTPUT_FILE_TEMP, ">$outputFileNameTemp")==0 ) {
	    $MONITOR && do { 
		print STDERR ("\nI cannot write/overwrite $outputFileNameTemp!\n");
		print STDERR ("I will stop processing this one!\a\n"); };
	    close(OUTPUT_FILE_TEMP);
	} elsif (open(OUTPUT_FILE, ">$outputFileName")==0) {
	    $MONITOR && do{
		print STDERR ("\nI cannot write/overwrite $outputFileName!\n");
		print STDERR ("I will stop processing this one!\a\n"); };
	    close(OUTPUT_FILE);
	} else {
	    # close first
	    close(OUTPUT_FILE_TEMP);
	    close(OUTPUT_FILE);
	    
	    $typeDo=$extension;
	    $typeDo=~tr/a-z/A-Z/;

	    $options = "-u \"http:\/\/$blastServerURL\" -f PROGRAM=$program DATALIB=$database INPUT_TYPE=\"sequence in fasta format\" SEQUENCE=\"$sequenceInFastaFormat\" EXPECT=$expect FILTER=$filter $NCBI_giOption DESCRIPTIONS=$descriptions ALIGNMENTS=$alignments $graphicalOverviewOption GENETICS_CODE=$queryGeneticCode OTHER_ADVANCED=\"$advancedOptions\" MATRIX=$matrix";

	    $version eq '1'   and $options .= " CUTOFF=default";
	    $version eq '2'   and $options .= " $gappedAlignmentOption";
	    $version eq 'psi' and $options .= " E_THRESH=$expectPSI";
	    
#	    print STDERR "\nversion = $version\noptions: \n$options";<STDIN>;

	    $MONITOR && print STDERR "Sending sequence $id to blast server.\n(Prog=$typeDo, Version=$version, DB=$database, Return=$resultReturnProtocol)\n";

	    # if return via www
	    if ($resultReturnProtocol eq "www") {
		my ($outFile);

		$MONITOR && print STDERR "Be patient, especially during peak hours :-)...\n";
		# Run the Blast:

		system("$pathOfPostclient $options > $outputFileNameTemp");


#		$MONITOR && print STDERR "Cleaning the file a bit...\n";
		if (&_removeJunkTagAndText($outputFileNameTemp, $outputFileName)==1) {
		    unlink $outputFileNameTemp;
		    $outFile = $outputFileName;
		    $MONITOR && print STDERR "$typeDo result saved to file $outputFileName\n";

		    # If html flag is off, then call _removeHTMLtags
		    # If the report is to be parsed, the non-HTML version must be saved
		    # in @_outFiles.
		    if ($resultTypeInHTMLflag=~m/^off$/i || $resultTypeInHTMLflag =~ m/^both$/i ) {
			$outFile = &_removeHTMLtags($outputFileName, $baseFileName);
		    }
		    $countSuccessfulBlasting++;
		} else {
		    # SAC: (Note) _removeJunkTagAndText() failed. 
		    #      Going with the temp file. Repercussions?
		    unlink $outputFileName;
		    if(-s $outputFileNameTemp) {
			$outFile = $outputFileNameTemp;
			$MONITOR && print STDERR "Possibly bad Blast result saved to file $outputFileNameTemp\n"; 
		    } else {
			unlink $outputFileNameTemp;
			$MONITOR && print STDERR "No Blast results saved.\n";
		    }
		}

		# SAC: Save the name of the Blast output file.
		push @_outFiles, $outFile;

	    }  else {  # by email (no results file to process.)
		
		$MONITOR && print STDERR "Results will be sent via e-mail: $emailAddress.\n";
		$options .= " EMAIL=is_set PATH=$emailAddress $resultTypeOption";
		# Run the Blast:
		system("$pathOfPostclient $options > $outputFileNameTemp");
		&_removeJunkTagAndTextForEmailResponse($outputFileNameTemp);
		$countSuccessfulBlasting++;
		push @_outFiles, 'email';
	    }
	}
	if ($_errmsg and $count < scalar(@queue)) {
	    print STDERR "\n$_errmsg\n";
	    $_errmsg = '';
	}
    } # end foreach $seq()
    
    if(not $countSuccessfulBlasting) {
	$_errmsg ||= "Blast failed: No file created or a error occured upon submission.";
	$_blastObj->throw($_errmsg);
    }

    if ($resultReturnProtocol eq "www") {
	$MONITOR && print STDERR "\n$countSuccessfulBlasting file(s) have been ${typeDo}ed successfully!\n\n";
    } else {
	$MONITOR && print STDERR "$countSuccessfulBlasting file(s) have been sent to blast server successfully\n\n";
    }
#    $MONITOR && print STDERR "All done!\a\n\n";
}


#------------------
sub _validate_seq {
#------------------
# SAC: sequence error checking using methods on the sequence object.
# No longer need Alex's various file checking functions since we are
# obtaining the data directly from sequence objects.
# Throws  : Exception if the sequence type is incompatible with the 
#           version of Blast to be run or the database selected.
# Returns : zero if sequence length is too short or too long
#           otherwise returns 1.
    my $seq = shift;
    my $type = $seq->type;
    my $len = $seq->seq_len;

    # Verify that sequence type is correct type for selected program.
    if(($type =~ /[dr]na/i and ($program =~ m/^(blastp|tblastn)$/i))
	  or
       ($type =~ /amino/i and ($program =~ m/^(blast[nx]|tblastx|dbest)$/i))
       ) {
	$_blastObj->throw("The sequence of ${\$seq->id} is wrong type for program $program");
    }

    # Verify that sequence type is correct type for selected database.
    if(($type =~ /[dr]na/i and not (grep $database eq $_, @Blast_dbn_remote))
       or
       ($type =~ /amino/i and not (grep $database eq $_, @Blast_dbp_remote))
       ) {
	$_blastObj->throw("The sequence of ${\$seq->id} is wrong type for database $database");
    }

    # SAC: added more detailed error message.
    if ($len < $minimumLength) {
	$MONITOR && print STDERR ("\nThe sequence length of ${\$seq->id} ($len) too short for blasting: min length = $minimumLength\a\n\n");
	return 0;
    } elsif( $len > $maximumLength) {
	$MONITOR && print STDERR ("\nThe sequence length of ${\$seq->id} ($len) is too long for blasting: max length = $maximumLength\a\n\n");
	return 0;
    }
    1;
}



#--------------------------
sub _removeJunkTagAndText {
#--------------------------
# Purpose : removes junk tags+text from the file, and also retrieve .gif files if graphical is on
# also change relative URL of entries to absolute path.
# return 0 and if any error! or 1 if success
#
# SAC: Added some blocks to detect server errors.
#      Cleaned up some regexps.
#      Reduced the dependency on NCBI output format.
#      Using class-scope private variable to pass error msg back to caller.

   my ($inputFileName, $outputFileName)=@_;
   my ($line, $count);

   # test if files are 0 size!
   if (!-s $inputFileName) {
       $_errmsg = "File $inputFileName is empty! Possible reasons: \nblast server is too busy".
	   " or postclient.pl does not work.";
      return 0;
   }

   # test if files are readable and writable
   if (open(INPUT_FILE_FOR_REMOVE, $inputFileName)==0) {
      $_errmsg = "File $inputFileName is not readable! Can't process it.";
      close(INPUT_FILE_FOR_REMOVE);
      return 0;
   }
   # the following file is writable since I have tested already
   open(OUTPUT_FILE_FOR_REMOVE, ">$outputFileName");

   print OUTPUT_FILE_FOR_REMOVE "<HTML><PRE>\n";

   ## BEGIN EXTRACTING REPORT DATA.

   # ignore header until the name/version of program BLASTN 2.0.3 [Nov-14-1997]
   while (1) {
      $line=<INPUT_FILE_FOR_REMOVE>;

      # SAC: New server error detection block.
      if($line =~ /Error \d+ .+?/ or
	 $line =~ /Internal Server Error/i)  {
	  my $msg = $&;
	  my @remainder = <INPUT_FILE_FOR_REMOVE>;
	  $_errmsg = $1."\n".join(' ',@remainder);
	  return 0;
      }

      # SAC: more general purpose regexp for program line.
      last if $line =~ m/T?BLAST[NPX]\s[\w.-]/;

      if (eof(INPUT_FILE_FOR_REMOVE)) {
          close(INPUT_FILE_FOR_REMOVE);
          close(OUTPUT_FILE_FOR_REMOVE);
          $_errmsg = "Blast file $inputFileName is incomplete: can't find program name.".
	             "\nPossibly an unrecognized report format." ;
          return 0;
      }
   }

   # copy BLASTN 2.0.3 [Nov-14-1997]
   print OUTPUT_FILE_FOR_REMOVE "$line\n";

   # add date/time stamp:
   print OUTPUT_FILE_FOR_REMOVE "\n<b>Done at $currentDate (local time)</b>\n";

   # SAC: The following commented out section is from the original webblast which
   # skips the "Reference:" text. It should be kept for documentation purposes.
   # Also, this code is sensitive to changes in the output format.

   # ignore until Query=   
#   while (1) {
#      $line=<INPUT_FILE_FOR_REMOVE>;
#      if ($line=~m/Query=/) {
#	   last;
#      }
#      if (eof(INPUT_FILE_FOR_REMOVE)) {
#	   close(INPUT_FILE_FOR_REMOVE);
#	   close(OUTPUT_FILE_FOR_REMOVE);
#	   $_errmsg = "Blast file $inputFileName was not completed!".
#		      "\nPossibly an unrecognized report format (2)." ;
#	   return 0;
#      }
#   }
#   # copy Query= line
#   print OUTPUT_FILE_FOR_REMOVE $line;
#
#   # copy next 6 lines, adding time stamp
#   for ($count=0; $count<6; $count++) {
#      $line=<INPUT_FILE_FOR_REMOVE>;
#      if (eof(INPUT_FILE_FOR_REMOVE)) {
#	    close(INPUT_FILE_FOR_REMOVE);
#	    close(OUTPUT_FILE_FOR_REMOVE);
#	    $_errmsg = "Blast file $inputFileName was not completed!".
#		       "\nPossibly an unrecognized report format (3)." ;
#	    return 0;
#      }
#      # add time to this line:
#      $line =~ s/^(Searching.+)$/$1\ndone at $currentDate (local time)/;
#
#      print OUTPUT_FILE_FOR_REMOVE $line;
#   }

   # copy all lines, adding absolute URL stem to all links:
   while (!eof(INPUT_FILE_FOR_REMOVE)) {
      $line=<INPUT_FILE_FOR_REMOVE>;
      # SAC: substituting in one step.
      $line =~ s|<a href="/|<a href="http://$entryServerURLroot/|;
      print OUTPUT_FILE_FOR_REMOVE $line;
   }
   close(INPUT_FILE_FOR_REMOVE);
   close(OUTPUT_FILE_FOR_REMOVE);

   return 1;
}



#--------------------------------------------
sub _removeJunkTagAndTextForEmailResponse {
#--------------------------------------------
# Purpose :b removes junk tags+text from the file, and also retrieve .gif files if graphical is on
# return 0 if any error! or 1 if success

   my ($inputFileName)=@_;
   my ($line);

   # test if files are readable and writable
   if (open(INPUT_FILE_FOR_REMOVE, $inputFileName)==0) {
      close(INPUT_FILE_FOR_REMOVE);
      $_errmsg = "Email response file $inputFileName is not readable! Can't process it.";
      return 0;
   }

   print "\n";
   # copy lines until cgi-bin/imagemap/BLAST/blast_form.map
   while (!eof(INPUT_FILE_FOR_REMOVE)) {
      $line=<INPUT_FILE_FOR_REMOVE>;
      $line=~s/<[^>]+>//g;
      print "$line";
   }

   close(INPUT_FILE_FOR_REMOVE);
   unlink $inputFileName;

   return 1;
}


#----------------------
sub _removeHTMLtags {
#----------------------
# Purpose : removes all html tags!
# return 0 if any error or 1 if success
# SAC:
#   Modifications:
#    - Uses Bio::Tools::Blast::HTML::strip_html() to remove HTML.
#      Stripping now occurs on the whole report in one chunk instead
#      of line-by-line (line-by-line is more error-prone).
#    - Now returns the name of the new file (or zero on failure).
#    - adding time stamp, as was done in _removeJunkTagAndText().
#    - cleaned up the HTML removal code a bit
#    - fixed bug due to failure to close the filehandle
#      (final close calls were missing '_TAG'.
#    - Moved the actual HTML-removing code into Bio::Tools::HTML
#
#    Note that $inputFileName is the ".program.html" file
#    which is deleted by this function if not saving the HTML version. 

   my ($inputFileName, $baseFileName )=@_;
   my ($line, $lineWithTag, $fileNameTagRemoved);
#   $fileNameTagRemoved  = $baseFileName . ".$extension" . ".v2";
   # SAC: don't need '.v2' since it was saved with a .html extension.
   $fileNameTagRemoved   = $baseFileName;

   # test if files are 0 size!
   if (!-s $inputFileName) {
      $_errmsg = "File $inputFileName is empty! Can't remove HTML tags.";
      return 0;
   }

   # test if temp file exist already or not writable
   if (-e $fileNameTagRemoved && !-w $fileNameTagRemoved) {
      $_errmsg = "Can't create temp file $fileNameTagRemoved! Can't process $inputFileName.";
      return 0;
   }

#   $MONITOR && print STDERR ("removing all html tags...\n");
   open(INPUT_FILE_FOR_REMOVE_TAG, "$inputFileName");
   open(OUTPUT_FILE_FOR_REMOVE_TAG, ">$fileNameTagRemoved");

   while (1) {
         $line=<INPUT_FILE_FOR_REMOVE_TAG>;
	 # SAC: more general purpose regexp for program line.
	 if ($line =~ m/T?BLAST[NPX]\s[\w.-]/) {
#         if ($line=~m/[a-z]{5,7} \d+\.\d+\.\d+ \[[a-z]{3}-\d{2}-\d{4}\]/i) {
             # remove tags, print to output, then exit loop
	     $line =~ s/<[^>]+>//g;
             print OUTPUT_FILE_FOR_REMOVE_TAG $line;
             last;
         }

         if (eof(INPUT_FILE_FOR_REMOVE_TAG)) {
             $_errmsg = "Blast file $inputFileName is incomplete: Can't find program name.";
             close(INPUT_FILE_FOR_REMOVE_TAG);
             close(OUTPUT_FILE_FOR_REMOVE_TAG);
             return 0;
         }
    }

   # removing html tags
   # SAC: Slurping the remainder of the file and stripping the whole thing at once
   #      using Bio::Tools::Blast::HTML::strip_html().
   my @lines = <INPUT_FILE_FOR_REMOVE_TAG>;
   $line = join('', @lines);
   &strip_html(\$line); 
   print OUTPUT_FILE_FOR_REMOVE_TAG $line;

   close(INPUT_FILE_FOR_REMOVE_TAG);
   close(OUTPUT_FILE_FOR_REMOVE_TAG);

   $MONITOR && print STDERR ("Output without HTML tags saved to $fileNameTagRemoved\n");

   # now if not ON or BOTH!
   # remove the .html file.
   if ($resultTypeInHTMLflag=~m/^off$/i) {
      unlink $inputFileName; 
      $MONITOR && print STDERR ("Deleted HTML file.\n");
  } 

   
   return $fileNameTagRemoved;
}



1;
__END__

#####################################################################################
#                                END OF CLASS                                       #
#####################################################################################

